/* This file is part of mailfrom filter.
   Copyright (C) 2006 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <libmilter/mfapi.h>

void
ml_free(int mc, char **mv)
{
	int i;

	for (i = 0; i < mc; i++)
		free(mv[i]);
	free(mv);
}

int
ml_split(char *message, int *argc, char ***argv)
{
	char *p;
	int i, c;
	char **v;

	c = 0;
	for (p = message; *p; p++)
		if (*p == '\n')
			c++;
	v = calloc(c + 1, sizeof v[0]);
	for (i = 0, p = message; i < c; i++) {
		size_t len = strcspn(p, "\r\n");
		char *np = malloc(len + 1);

		if (!np) {
			ml_free(i, v);
			return 1;
		}
		memcpy(np, p, len);
		np[len] = 0;
		v[i] = np;
		p += len;
		if (*p == '\r')
			p++;
		p++;
	}
	v[i] = NULL;
	*argc = c;
	*argv = v;
	return 0;
}

/* The awfully ugly function below is needed because of the brain-
   damaged libmilter interface, which supplies only function smfi_setmlreply
   with a variable number of arguments. There's no way to pass it a newline-
   separated string, or a string array, or even a va_alist. Moreover,
   the number of lines in a reply is limited to 32!
*/
int
sendmail_mlreply(SMFICTX * ctx, const char *rcode, const char *xcode,
		 char *message)
{
	int rc;
	int mc;
	char **mv;

	if (ml_split(message, &mc, &mv))
		return 1;
	switch (mc) {
	case 1:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], NULL);
		break;
	case 2:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     NULL);
		break;
	case 3:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], NULL);
		break;
	case 4:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], NULL);
		break;
	case 5:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], NULL);
		break;
	case 6:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], NULL);
		break;
	case 7:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     NULL);
		break;
	case 8:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], NULL);
		break;
	case 9:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], NULL);
		break;
	case 10:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], NULL);
		break;
	case 11:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], NULL);
		break;
	case 12:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     NULL);
		break;
	case 13:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], NULL);
		break;
	case 14:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], NULL);
		break;
	case 15:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], NULL);
		break;
	case 16:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15], NULL);
		break;
	case 17:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], NULL);
		break;
	case 18:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], NULL);
		break;
	case 19:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], NULL);
		break;
	case 20:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19], NULL);
		break;
	case 21:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19],
				     mv[20], NULL);
		break;
	case 22:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19],
				     mv[20], mv[21], NULL);
		break;
	case 23:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19],
				     mv[20], mv[21], mv[22], NULL);
		break;
	case 24:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19],
				     mv[20], mv[21], mv[22], mv[23], NULL);
		break;
	case 25:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19],
				     mv[20], mv[21], mv[22], mv[23],
				     mv[24], NULL);
		break;
	case 26:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19],
				     mv[20], mv[21], mv[22], mv[23],
				     mv[24], mv[25], NULL);
		break;
	case 27:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19],
				     mv[20], mv[21], mv[22], mv[23],
				     mv[24], mv[25], mv[26], NULL);
		break;
	case 28:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19],
				     mv[20], mv[21], mv[22], mv[23],
				     mv[24], mv[25], mv[26], mv[27], NULL);
		break;
	case 29:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19],
				     mv[20], mv[21], mv[22], mv[23],
				     mv[24], mv[25], mv[26], mv[27],
				     mv[28], NULL);
		break;
	case 30:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19],
				     mv[20], mv[21], mv[22], mv[23],
				     mv[24], mv[25], mv[26], mv[27],
				     mv[28], mv[29], NULL);
		break;
	case 31:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19],
				     mv[20], mv[21], mv[22], mv[23],
				     mv[24], mv[25], mv[26], mv[27],
				     mv[28], mv[29], mv[30], NULL);
		break;
	case 32:
		rc = smfi_setmlreply(ctx, rcode, xcode, mv[0], mv[1],
				     mv[2], mv[3], mv[4], mv[5], mv[6],
				     mv[7], mv[8], mv[9], mv[10], mv[11],
				     mv[12], mv[13], mv[14], mv[15],
				     mv[16], mv[17], mv[18], mv[19],
				     mv[20], mv[21], mv[22], mv[23],
				     mv[24], mv[25], mv[26], mv[27],
				     mv[28], mv[29], mv[30], mv[31], NULL);
		break;
	default:
		rc = 1;
	}
	ml_free(mc, mv);
	return rc;
}
